package fr.lri.hcc.ilda.touchtokens;

/*
 *   Authors: Caroline Appert (caroline.appert@lri.fr)
 *   Copyright (c) Univ. Paris-Sud XI, 2016. All Rights Reserved
 *   Licensed under the GNU LGPL. For full terms see the file COPYING.
 */

import java.awt.geom.GeneralPath;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.text.DecimalFormat;
import java.text.DecimalFormatSymbols;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;

public class TouchTokenRecognizer {

	// Set to true if we want the recognizer to return the closest token in all cases
	// Set to false if we want the recognizer to return the closest token only if the input 3-point pattern falls within the tolerance area of the token's template
	public static boolean TOLERANT = false;

	protected ArrayList<TokenTemplate> tokenTemplates;

	public TouchTokenRecognizer() {
		this.tokenTemplates = new ArrayList<TokenTemplate>();
	}

	public TouchTokenRecognizer(File templatesFile) {
		this.tokenTemplates = new ArrayList<TokenTemplate>();
		try {
			BufferedReader br = new BufferedReader(new FileReader(templatesFile));
			String line = br.readLine();
			DecimalFormat df = new DecimalFormat();
			DecimalFormatSymbols symbols = new DecimalFormatSymbols();
			symbols.setDecimalSeparator('.');
			symbols.setGroupingSeparator(',');
			df.setDecimalFormatSymbols(symbols);
			double x, y;
			while (line != null) {
				//				System.out.println(line);
				if (!line.startsWith("#")) {
					String[] parts = line.split(",");
					if (parts.length > 3) {
						String tokenID = parts[0];
						int fingerCount = Integer.parseInt(parts[1]);
						ArrayList<TouchPoint> points = new ArrayList<TouchPoint>();
						for (int i = 0; i < fingerCount; i++) {
							x = df.parse(parts[2+(i*2)]).doubleValue();
							y = df.parse(parts[2+(i*2)+1]).doubleValue();
							TouchPoint point = new TouchPoint(x, y);
							points.add(point);
						}
						TouchPoint center = new TouchPoint(0, 0);
						if(parts.length > (2+(fingerCount*2))) {
							double centerx = df.parse(parts[2+(fingerCount*2)]).doubleValue();
							double centery = df.parse(parts[2+(fingerCount*2)+1]).doubleValue();
							center.setLocation(centerx, centery);
							System.out.println("CENTER provided ("+centerx+", "+centery+")");
						} else {
							System.out.println("NO CENTER provided");
						}

						int index = 2+(fingerCount*2)+2;
						boolean first = true;
						GeneralPath outline = new GeneralPath();
						while(index < parts.length) {
							x = df.parse(parts[index]).doubleValue();
							y = df.parse(parts[index+1]).doubleValue();
							if(first) {
								first = false;
								outline.moveTo(x, y);
							} else {
								outline.lineTo(x, y);
							}
							index += 2;
						}
						outline.closePath();

						TokenTemplate tt = new TokenTemplate(points, tokenID, center);
						tt.setContourShape(outline);
						tokenTemplates.add(tt);
					}
				}
				line = br.readLine();
			}
			br.close();
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		} catch (ParseException e) {
			e.printStackTrace();
		}

	}

	public void addTemplate(TokenTemplate tt) {
		tokenTemplates.add(tt);
	}

	public static void centerOnPoint(ArrayList<TouchPoint> points, TouchPoint refPoint) {
		for (TouchPoint pt : points) {
			pt.setLocation(pt.getX() - refPoint.getX(), pt.getY() - refPoint.getY());
		}
	}

	public static double angleBetweenVectors(TouchPoint vector1, TouchPoint vector2) {
		double angle = Math.atan2(vector2.getY(), vector2.getX()) - Math.atan2(vector1.getY(), vector1.getX());
		//	normalize to the range 0 .. 2 * Pi:
		if (angle < 0) angle += 2 * Math.PI;
		return angle;
	}

	public static void sortCCW(ArrayList<TouchPoint> points, final TouchPoint refPoint) {
		final TouchPoint vectorXAxis = new TouchPoint(10, 0);
		Collections.sort(points, new Comparator<TouchPoint>() {
			public int compare(TouchPoint pt1, TouchPoint pt2) {
				TouchPoint vector1 = new TouchPoint(pt1.getX() - refPoint.getX(), pt1.getY() - refPoint.getY());
				double angle1 = angleBetweenVectors(vectorXAxis, vector1);
				TouchPoint vector2 = new TouchPoint(pt2.getX() - refPoint.getX(), pt2.getY() - refPoint.getY());
				double angle2 = angleBetweenVectors(vectorXAxis, vector2);
				return (int)((angle2 - angle1)*100);
			}
		});
	}

	public static void rotateBy(ArrayList<TouchPoint> points, TouchPoint refPoint, double theta, ArrayList<TouchPoint> newPoints) {
		TouchPoint ptSrc, ptDest;
		for (int i = 0; i < points.size(); i++) {
			ptSrc = points.get(i);
			if (newPoints.size() > i) {
				ptDest = newPoints.get(i);
			} else {
				ptDest = new TouchPoint();
				newPoints.add(i, ptDest);
			}
			double x = (ptSrc.getX() - refPoint.getX()) * Math.cos(theta) - (ptSrc.getY() - refPoint.getY()) * Math.sin(theta) + refPoint.getX();
			double y = (ptSrc.getX() - refPoint.getX()) * Math.sin(theta) + (ptSrc.getY() - refPoint.getY()) * Math.cos(theta) + refPoint.getY();
			ptDest.setLocation(x, y);
		}
	}

	public static void rotateBy(TouchPoint point, TouchPoint refPoint, double theta, TouchPoint newPoint) {
		double x = (point.getX() - refPoint.getX()) * Math.cos(theta) - (point.getY() - refPoint.getY()) * Math.sin(theta) + refPoint.getX();
		double y = (point.getX() - refPoint.getX()) * Math.sin(theta) + (point.getY() - refPoint.getY()) * Math.cos(theta) + refPoint.getY();
		newPoint.setLocation(x, y);
	}

	public static double sortAndAlign(ArrayList<TouchPoint> points, TouchPoint refPoint, ArrayList<TouchPoint> resPoints) {
		resPoints.clear();
		for (TouchPoint pt : points) {
			resPoints.add(new TouchPoint(pt.getX(), pt.getY()));
		}
		centerOnPoint(resPoints, refPoint);
		TouchPoint origin = new TouchPoint(0, 0);
		sortCCW(resPoints, origin);
		TouchPoint refVector = new TouchPoint(resPoints.get(0).getX(), resPoints.get(0).getY());
		TouchPoint xAxisVector = new TouchPoint(10, 0);
		double angle = angleBetweenVectors(refVector, xAxisVector);
		rotateBy(resPoints, origin, angle, resPoints);
		return angle;
	}

	public static void sort(ArrayList<TouchPoint> points, TouchPoint refPoint, ArrayList<TouchPoint> resPoints) {
		resPoints.clear();
		for (TouchPoint pt : points) {
			resPoints.add(new TouchPoint(pt.getX(), pt.getY()));
		}
		centerOnPoint(resPoints, refPoint);
		TouchPoint origin = new TouchPoint(0, 0);
		sortCCW(resPoints, origin);
		//		TouchPoint refVector = new TouchPoint(resPoints.get(0).getX(), resPoints.get(0).getY());
		//		TouchPoint xAxisVector = new TouchPoint(10, 0);
		//		double angle = angleBetweenVectors(refVector, xAxisVector);
		//		rotateBy(resPoints, origin, angle, resPoints);
		//		return angle;
	}

	public static double sortAndAlignTemplate(TokenTemplate template, TouchPoint refPoint) {
		centerOnPoint(template.getPoints(), refPoint);
		template.getCenter().setLocation(template.getCenter().getX() - refPoint.x, template.getCenter().getY() - refPoint.y);

		TouchPoint origin = new TouchPoint(0, 0);
		sortCCW(template.getPoints(), origin);
		TouchPoint refVector = new TouchPoint(template.getPoints().get(0).getX(), template.getPoints().get(0).getY());
		TouchPoint xAxisVector = new TouchPoint(10, 0);
		double angle = angleBetweenVectors(refVector, xAxisVector);
		rotateBy(template.getPoints(), origin, angle, template.getPoints());


		double x = template.getCenter().getX() * Math.cos(angle) - template.getCenter().getY() * Math.sin(angle);
		double y = template.getCenter().getX() * Math.sin(angle) + template.getCenter().getY() * Math.cos(angle);
		template.getCenter().setLocation(x, y);

		return angle;
	}


	public static void rotateTemplate(TokenTemplate template, double angle) {
		ArrayList<TouchPoint> resPoints = new ArrayList<TouchPoint>();
		for (TouchPoint pt : template.points) {
			resPoints.add(new TouchPoint(pt.getX(), pt.getY()));
		}
		TouchPoint origin = new TouchPoint(0, 0);
		rotateBy(resPoints, origin, angle, resPoints);
		TouchPoint resCenter = new TouchPoint(template.getCenter().x, template.getCenter().y);
		double x = resCenter.getX() * Math.cos(angle) - resCenter.getY() * Math.sin(angle);
		double y = resCenter.getX() * Math.sin(angle) + resCenter.getY() * Math.cos(angle);
		resCenter.setLocation(x, y);
		template.setPoints(resPoints);
		template.setCenter(resCenter);
	}

	public static void translateTemplate(TokenTemplate template, TouchPoint translate) {
		for (TouchPoint pt : template.points) {
			pt.setLocation(pt.x + translate.x, pt.y + translate.y);
		}
		template.getCenter().setLocation(template.getCenter().x + translate.x, template.getCenter().y+translate.y);
	}

	public static TouchPoint centroid(ArrayList<TouchPoint> points) {
		double sumX = 0;
		double sumY = 0;
		for (Iterator<TouchPoint> iterator = points.iterator(); iterator.hasNext();) {
			TouchPoint next = iterator.next();
			sumX += next.getX();
			sumY += next.getY();
		}
		int length = points.size();
		return new TouchPoint(sumX / length, sumY / length);
	}

	public static TouchPoint getTokenCenter(TokenTemplate transformedTemplate, ArrayList<TouchPoint> notchPoints, double deviceWidth, double deviceHeight) {
		TouchPoint centerTemplate = new TouchPoint(transformedTemplate.getCenter().x, transformedTemplate.getCenter().y);
		TouchPoint pt1Template = new TouchPoint(transformedTemplate.getPoints().get(0).x, transformedTemplate.getPoints().get(0).y);
		TouchPoint pt2Template = new TouchPoint(transformedTemplate.getPoints().get(1).x, transformedTemplate.getPoints().get(1).y);
		TouchPoint v1CenterTemplate = new TouchPoint(centerTemplate.x - pt1Template.x, centerTemplate.y - pt1Template.y);
		TouchPoint v12Template = new TouchPoint(pt2Template.x - pt1Template.x, pt2Template.y - pt1Template.y);

		double angle21Center = TouchTokenRecognizer.angleBetweenVectors(v12Template, v1CenterTemplate);

		TouchPoint pt1Token = notchPoints.get(0);
		pt1Token = new TouchPoint(pt1Token.x*deviceWidth, pt1Token.y*deviceHeight);
		TouchPoint pt2Token = notchPoints.get(1);
		pt2Token = new TouchPoint(pt2Token.x*deviceWidth, pt2Token.y*deviceHeight);
		TouchPoint v12Token = new TouchPoint(pt2Token.x - pt1Token.x, pt2Token.y - pt1Token.y);
		double x12Token = TouchTokenRecognizer.angleBetweenVectors(new TouchPoint(1, 0), v12Token);

		double length1Center =
				Math.sqrt((centerTemplate.x - pt1Template.x) * (centerTemplate.x - pt1Template.x) +
						(centerTemplate.y - pt1Template.y) * (centerTemplate.y - pt1Template.y));

		TouchPoint centerToken = new TouchPoint(length1Center, 0);
		centerToken.setLocation(centerToken.x + pt1Token.getX(), centerToken.y + pt1Token.getY());
		double angle = x12Token + angle21Center;
		TouchTokenRecognizer.rotateBy(centerToken, pt1Token, angle, centerToken);
		centerToken.setLocation(centerToken.x/deviceWidth, centerToken.y/deviceHeight);
		return centerToken;
	}

	public static ArrayList<ArrayList<TouchPoint>> allPermutations(ArrayList<TouchPoint> points) {
		ArrayList<ArrayList<TouchPoint>> allPermutations = new ArrayList<ArrayList<TouchPoint>>();
		for (int i = 0; i < points.size(); i++) {
			ArrayList<TouchPoint> order = new ArrayList<TouchPoint>();
			order.add(points.get(i));
			for(int j = i+1; j < points.size(); j++) {
				order.add(points.get(j));
			}
			for(int j = 0; j < i; j++) {
				order.add(points.get(j));
			}
			allPermutations.add(order);
		}
		return allPermutations;
	}

	public static double meanSquaredPairedPointDistanceForAlignedPairs(ArrayList<TouchPoint> points1, ArrayList<TouchPoint> points2) {
		if(points1.size() != points2.size()) {
			return Double.MAX_VALUE;
		} else {
			double res = 0;
			for (int i = 0; i < points1.size(); i++) {
				TouchPoint p1 = points1.get(i);
				TouchPoint p2 = points2.get(i);
				res += (p1.getX() - p2.getX()) * (p1.getX() - p2.getX()) + (p1.getY() - p2.getY()) * (p1.getY() - p2.getY());
			}
			return res / points1.size();
		}
	}

	public static DistanceResult minDistance(TokenTemplate templateOrderedAndAligned, ArrayList<TouchPoint> inputPoints, TouchPoint refPoint) {
		ArrayList<TouchPoint> inputPointsOrderedAndAligned = new ArrayList<TouchPoint>();
		//		double alignAngle = sortAndAlign(inputPoints, refPoint, inputPointsOrderedAndAligned);
		sort(inputPoints, refPoint, inputPointsOrderedAndAligned);

		ArrayList<ArrayList<TouchPoint>> allPermutations = allPermutations(inputPointsOrderedAndAligned);
		double minDis = Double.MAX_VALUE;
		double angleRotate = 0;
		for (ArrayList<TouchPoint> inputPermutation : allPermutations) {
			TouchPoint origin = new TouchPoint(0, 0);
			TouchPoint refVector = new TouchPoint(inputPermutation.get(0).getX(), inputPermutation.get(0).getY());
			TouchPoint xAxisVector = new TouchPoint(10, 0);
			double angle = angleBetweenVectors(refVector, xAxisVector);
			ArrayList<TouchPoint> inputPermutationRotated = new ArrayList<TouchPoint>();
			rotateBy(inputPermutation, origin, angle, inputPermutationRotated);
			double d = meanSquaredPairedPointDistanceForAlignedPairs(inputPermutationRotated, templateOrderedAndAligned.getPoints());
			if(d < minDis) {
				minDis = d;
				angleRotate = angle;
			}
		}

		// modify the template to align it with input
		//		rotateTemplate(templateOrderedAndAligned, -(angleRotate+alignAngle));
		rotateTemplate(templateOrderedAndAligned, -(angleRotate));
		translateTemplate(templateOrderedAndAligned, new TouchPoint(refPoint.x, refPoint.y));

		return new DistanceResult(angleRotate, minDis);
	}

	public TokenTemplate sortTemplatePoints(ArrayList<TouchPoint> input, TokenTemplate tokenTemplate) {
		if(input.size() != 3) {
			return null;
		}
		ArrayList<TouchPoint> points = new ArrayList<TouchPoint>();
		for (TouchPoint touchPoint : tokenTemplate.getPoints()) {
			points.add(new TouchPoint(touchPoint.x, touchPoint.y));
		}
		TouchPoint center = new TouchPoint(tokenTemplate.getCenter().x, tokenTemplate.getCenter().y);
		TokenTemplate templateTransformed = new TokenTemplate(points, tokenTemplate.getTokenID(), center);
		double templateAngle = sortAndAlignTemplate(templateTransformed, centroid(tokenTemplate.getPoints()));
		DistanceResult inputDistance = minDistance(templateTransformed, input, centroid(input));
		templateTransformed.setRecognitionDistance(inputDistance.getDistance());
		templateTransformed.setOrientation(inputDistance.getAngle()-templateAngle);

		// sort template points
		ArrayList<TouchPoint> unsortedTemplatePoints = new ArrayList<TouchPoint>();
		unsortedTemplatePoints.addAll(templateTransformed.getPoints());
		ArrayList<TouchPoint> sortedTemplatePoints = new ArrayList<TouchPoint>();
		for(int i = 0; i < input.size(); i++) {
			double minDistance = Double.MAX_VALUE;
			int indexMinDistance = -1;
			for(int j = 0; j < unsortedTemplatePoints.size(); j++) {
				double d = input.get(i).distance(unsortedTemplatePoints.get(j));
				if(d < minDistance) {
					minDistance = d;
					indexMinDistance = j;
				}
			}
			TouchPoint point = unsortedTemplatePoints.remove(indexMinDistance);
			sortedTemplatePoints.add(point);
		}
		templateTransformed.setPoints(sortedTemplatePoints);
		return templateTransformed;
	}
	
	public TokenTemplate recognize(ArrayList<TouchPoint> input) {
		if(input.size() != 3) {
			return null;
		}
		for (TokenTemplate tokenTemplate : tokenTemplates) {
			tokenTemplate.setRecognitionDistance(Double.MAX_VALUE);
		}
		double minDis = Double.MAX_VALUE;
		TokenTemplate tokenRecognized = null;
		for (TokenTemplate tokenTemplate : tokenTemplates) {

			ArrayList<TouchPoint> points = new ArrayList<TouchPoint>();
			for (TouchPoint touchPoint : tokenTemplate.getPoints()) {
				points.add(new TouchPoint(touchPoint.x, touchPoint.y));
			}
			TouchPoint center = new TouchPoint(tokenTemplate.getCenter().x, tokenTemplate.getCenter().y);
			TokenTemplate templateTransformed = new TokenTemplate(points, tokenTemplate.getTokenID(), center);
//			templateTransformed.setOriginalGeometryTemplate(templateTransformed);
			double templateAngle = sortAndAlignTemplate(templateTransformed, centroid(tokenTemplate.getPoints()));
			//			double d = minDistance(templateTransformed, input, centroid(input));
			//			templateTransformed.setDistance(d);
			DistanceResult inputDistance = minDistance(templateTransformed, input, centroid(input));
			templateTransformed.setRecognitionDistance(inputDistance.getDistance());
			//			System.out.println("\t"+tokenTemplate.tokenID+" _ template:"+templateAngle+" _ input:"+inputDistance.getAngle());
			templateTransformed.setOrientation(inputDistance.getAngle()-templateAngle);
			//			templateTransformed.setAngle(dr.getAngle());

			if(inputDistance.getDistance() < minDis) {
				minDis = inputDistance.getDistance();
				tokenRecognized = templateTransformed;
			}
		}

		if(tokenRecognized == null) {
			return null;
		}

		// sort template points
		ArrayList<TouchPoint> unsortedTemplatePoints = new ArrayList<TouchPoint>();
		unsortedTemplatePoints.addAll(tokenRecognized.getPoints());
		ArrayList<TouchPoint> sortedTemplatePoints = new ArrayList<TouchPoint>();
		for(int i = 0; i < input.size(); i++) {
			double minDistance = Double.MAX_VALUE;
			int indexMinDistance = -1;
			for(int j = 0; j < unsortedTemplatePoints.size(); j++) {
				double d = input.get(i).distance(unsortedTemplatePoints.get(j));
				if(d < minDistance) {
					minDistance = d;
					indexMinDistance = j;
				}
			}
			TouchPoint point = unsortedTemplatePoints.remove(indexMinDistance);
			sortedTemplatePoints.add(point);
		}
		tokenRecognized.setPoints(sortedTemplatePoints);


		if(!TOLERANT) {
			// NEW begin
			// Test that input falls within the tolerance area of the template (i.e., each point is at a distance lesser than 5mm)

			double d1 = tokenRecognized.getPoints().get(0).distance(input.get(0));
			double d2 = tokenRecognized.getPoints().get(1).distance(input.get(1));
			double d3 = tokenRecognized.getPoints().get(2).distance(input.get(2));

			if(d1 <= 5 && d2 <= 5 && d3 <= 5) {
				return tokenRecognized;
			} else {
				return null;
			}

			// NEW end
		} else {
			return tokenRecognized;
		}
	}

	public static void main(String[] args) {
		new TouchTokenRecognizer(new File("templates/templates.txt"));
	}

	public ArrayList<TokenTemplate> getTokenTemplates() {
		return tokenTemplates;
	}

	public void setTokenTemplates(ArrayList<TokenTemplate> tokenTemplates) {
		this.tokenTemplates = tokenTemplates;
	}

	public TokenTemplate getTemplate(String tokenID) {
		for (TokenTemplate tokenTemplate : tokenTemplates) {
			if(tokenTemplate.getTokenID().compareTo(tokenID) == 0) {
				return tokenTemplate;
			}
		}
		return null;
	}

}
